"""DID Doc Service."""

from voluptuous import ALLOW_EXTRA, All, Schema, Switch, Url

from ..did_url import DIDUrl
from ..validation import wrap_validation_error, Into
from . import DIDDocumentError


class ServiceValidationError(DIDDocumentError):
    """Raised on service validation failure."""


class Service:
    """Representation of DID Document Services."""

    _validator = Schema(
        {
            "id": All(str, DIDUrl.validate),
            "type": str,
            "serviceEndpoint": Switch(DIDUrl.validate, Url(), ""),
        },
        extra=ALLOW_EXTRA,
        required=True,
    )

    def __init__(self, id_: DIDUrl, type_: str, endpoint: str, **extra):
        """Initialize Service."""
        self._id = id_
        self._type = type_
        self._endpoint = endpoint
        self._extra = extra

    @property
    def id(self):
        """Return id."""
        return self._id

    @property
    def type(self):
        """Return type."""
        return self._type

    @property
    def endpoint(self):
        """Return endpoint."""
        return self._endpoint

    @property
    def extra(self):
        """Return extra."""
        return self._extra

    def serialize(self):
        """Return serialized representation of Service."""
        return {
            "id": str(self.id),
            "type": self.type,
            "serviceEndpoint": self.endpoint,
            **self.extra,
        }

    @classmethod
    @wrap_validation_error(ServiceValidationError, message="Failed to validate service")
    def validate(cls, value: dict):
        """Validate object against service."""
        return cls._validator(value)

    @classmethod
    @wrap_validation_error(
        ServiceValidationError, message="Failed to deserialize service"
    )
    def deserialize(cls, value: dict):
        """Deserialize into Service."""
        value = cls.validate(value)
        deserializer = Schema(
            {
                Into("id", "id_"): DIDUrl.parse,
                Into("type", "type_"): str,
                Into("serviceEndpoint", "endpoint"): str,
            },
            extra=ALLOW_EXTRA,
        )
        value = deserializer(value)
        return cls(**value)



"""DID Doc Verification Method."""

from typing import Any, Set

from voluptuous import ALLOW_EXTRA, All, Coerce, Invalid, Remove, Schema, Union

from ..did import DID
from ..did_url import DIDUrl
from ..validation import Into, validate_init, wrap_validation_error
from . import DIDDocumentError
from .verification_method_options import VerificationMethodOptions


class InvalidVerificationMaterial(DIDDocumentError, Invalid):
    """Error raised when verification material is invalid."""


class VerificationMethodValidationError(DIDDocumentError):
    """Raised on VerificationMethod validation failure."""


def verification_material(value):
    """Validate Verification Material key."""
    if "publicKey" not in value:
        raise InvalidVerificationMaterial("Invalid")
    return value


class VerificationSuite:
    """
    Verification Suite of a Verification Method, defining type and material property.
    """

    def __init__(self, type_: str, verification_material_prop: str):
        """Initialize Suite.

        Args:
            type_ (str): type of the suite.
            verification_material_prop (str): property of the verification
                method map containing the key material.
        """
        self._type = type_
        self._verification_material_prop = verification_material_prop

    def __eq__(self, other):
        if not isinstance(other, VerificationSuite):
            return False
        return (
            self._type == other._type
            and self._verification_material_prop == other._verification_material_prop
        )

    @property
    def type(self):
        """Return type."""
        return self._type

    @property
    def verification_material_prop(self):
        """Return verification_material_prop."""
        return self._verification_material_prop

    @classmethod
    def derive(cls, type_: str, **potential_material_prop):
        """
        Derive the verification suite from a type string and a set of
        potential material properties.
        """
        for key in potential_material_prop:
            if "publicKey" in key:
                return cls(type_, key)
        raise InvalidVerificationMaterial(
            "the verification suite and material could not be derived from inputs"
        )


class VerificationMethod:
    """Representation of DID Document Verification Methods."""

    _validator = Schema(
        {
            "id": All(str, DIDUrl.validate),
            "type": str,
            "controller": All(str, DID.validate),
            All(str, verification_material): Union(str, dict),
        },
        required=True,
        extra=ALLOW_EXTRA,
    )

    @validate_init(id_=DIDUrl, suite=VerificationSuite, controller=DID)
    def __init__(
        self,
        id_: DIDUrl,
        suite: VerificationSuite,
        controller: DID,
        material: Any,
        **extra
    ):
        """Initialize VerificationMethod."""
        self._id = id_
        self._suite = suite
        self._controller = controller
        self._material = material
        self.extra = extra

    def __eq__(self, other):
        """Test equality."""
        if not isinstance(other, VerificationMethod):
            return False
        return (
            self._id == other._id
            and self._suite == other._suite
            and self._controller == other._controller
            and self._material == other._material
            and self.extra == other.extra
        )

    # pylint: disable=invalid-name
    @property
    def id(self):
        """Return id."""
        return self._id

    @property
    def suite(self):
        """Return suite."""
        return self._suite

    @property
    def type(self):
        """Return suite type."""
        return self._suite.type

    @property
    def controller(self):
        """Return controller."""
        return self._controller

    @property
    def material(self):
        """Return material."""
        return self._material

    def serialize(self):
        """Return serialized representation of VerificationMethod."""
        return {
            "id": str(self.id),
            "type": self.type,
            "controller": str(self.controller),
            self.suite.verification_material_prop: self.material,
            **self.extra,
        }

    @classmethod
    @wrap_validation_error(
        VerificationMethodValidationError,
        message="Failed to validate verification method",
    )
    def validate(cls, value: dict):
        """Validate against verification method."""
        return cls._validator(value)

    @classmethod
    @wrap_validation_error(
        VerificationMethodValidationError,
        message="Failed to deserialize verification method",
    )
    def deserialize(
        cls, value: dict, *, options: Set[VerificationMethodOptions] = None
    ):
        """Deserialize into VerificationMethod."""
        # Apply options
        if options:
            value = VerificationMethodOptions.apply(value, options)

        # Perform validation
        value = cls.validate(value)

        def _suite_and_material(value):
            suite = VerificationSuite.derive(value["type"], **value)
            material = value[suite.verification_material_prop]
            value["suite"] = suite
            value["material"] = material
            return value

        deserializer = Schema(
            All(
                {
                    Into("id", "id_"): All(str, DIDUrl.parse),
                    "controller": All(str, Coerce(DID)),
                },
                _suite_and_material,
                {Remove("type"): str, Remove(verification_material): Union(str, dict)},
            ),
            extra=ALLOW_EXTRA,
        )

        # Hydrate object
        value = deserializer(value)
        return cls(**value)



"""Service type for DID Communication Services."""

from typing import List

from voluptuous import PREVENT_EXTRA, All, Any, Schema, Coerce

from ..did_url import DIDUrl
from ..validation import wrap_validation_error, Into
from .service import Service, ServiceValidationError


class DIDCommService(Service):
    """DID Communication Service."""

    _validator = Service._validator.extend(
        {
            "type": Any("IndyAgent", "did-communication"),
            "recipientKeys": [All(str, DIDUrl.validate)],
            "routingKeys": [All(str, DIDUrl.validate)],
            "priority": int,
        },
        extra=PREVENT_EXTRA,
    )

    def __init__(
        self,
        id_: DIDUrl,
        endpoint: str,
        recipient_keys: List[DIDUrl],
        priority: int,
        *,
        type_: str = None,
        routing_keys: List[DIDUrl] = None
    ):
        """Initialize DIDCommService."""
        self._recipient_keys = recipient_keys
        self._routing_keys = routing_keys or []
        self.priority = priority
        super().__init__(
            id_,
            type_ or "did-communication",
            endpoint,
            recipient_keys=recipient_keys,
            routing_keys=routing_keys,
            priority=priority,
        )

    @property
    def recipient_keys(self):
        """Return recipient_keys."""
        return self._recipient_keys

    @property
    def routing_keys(self):
        """Return routing_keys."""
        return self._routing_keys

    def serialize(self):
        """Return serialized representation of DIDCommService."""
        did_urls = Schema([Coerce(str)])
        return {
            "id": str(self.id),
            "type": self.type,
            "serviceEndpoint": self.endpoint,
            "recipientKeys": did_urls(self.recipient_keys),
            "routingKeys": did_urls(self.routing_keys),
            "priority": self.priority,
        }

    @classmethod
    @wrap_validation_error(
        ServiceValidationError, message="Failed to validate DIDComm service"
    )
    def validate(cls, value: dict):
        """Validate value against DIDCommService."""
        return cls._validator(value)

    @classmethod
    @wrap_validation_error(
        ServiceValidationError, message="Failed to deserialize DIDComm service"
    )
    def deserialize(cls, value: dict):
        """Deserialize into Service."""
        value = cls.validate(value)
        deserializer = Schema(
            {
                Into("id", "id_"): DIDUrl.parse,
                Into("type", "type_"): str,
                Into("serviceEndpoint", "endpoint"): str,
                Into("recipientKeys", "recipient_keys"): [DIDUrl.parse],
                Into("routingKeys", "routing_keys"): [DIDUrl.parse],
                "priority": int,
            }
        )
        value = deserializer(value)
        return cls(**value)



"""DID Document and resource builders."""

from contextlib import contextmanager
from typing import Any, List, Union, Iterator

from voluptuous import All, Coerce, Switch

from ..did import DID
from ..did_url import DIDUrl
from ..validation import validate_init
from .didcomm_service import DIDCommService
from .doc import DIDDocument
from .service import Service
from .verification_method import VerificationMethod, VerificationSuite
from .verification_relationship import VerificationRelationship


def _default_id_generator(base: str, start: int = 0) -> Iterator[str]:
    """Generate ID fragments."""
    index = start
    while True:
        yield "{}-{}".format(base, index)
        index += 1


class VerificationMethodBuilder:
    """VerificationMethod scoped builder."""

    def __init__(
        self, did: DID, *, id_base: str = None, methods: List[VerificationMethod] = None
    ):
        self._did = did
        self.methods = methods or []
        self._id_base = id_base or "keys"
        self._id_generator = None
        self._default_suite = None

    def _default_id_generator(self):
        """Default ID generator."""
        yield from _default_id_generator(self._id_base, start=len(self.methods))

    @contextmanager
    def defaults(
        self, id_generator: Iterator[str] = None, suite: VerificationSuite = None
    ) -> Iterator["VerificationMethodBuilder"]:
        """Enter context with defaults set."""
        self._id_generator = id_generator or self._default_id_generator()
        self._default_suite = suite
        try:
            yield self
        finally:
            self._id_generator = None
            self._default_suite = None

    def add(
        self,
        material: Any,
        suite: VerificationSuite = None,
        ident: str = None,
        controller: DID = None,
    ):
        """Add verification method from parts and context."""
        if not ident and not self._id_generator:
            raise ValueError("No ident provided for method")
        ident = ident or next(self._id_generator)
        suite = suite or self._default_suite
        if not suite:
            raise ValueError("No VerificationSuite or default suite")
        controller = controller or self._did
        vmethod = VerificationMethod(
            id_=self._did.ref(ident),
            suite=suite,
            controller=controller,
            material=material,
        )
        self.methods.append(vmethod)
        return vmethod

    def remove(self, vmethod: VerificationMethod):
        """Remove method from builder."""
        self.methods.remove(vmethod)


class RelationshipBuilder(VerificationMethodBuilder):
    """Builder for relationships."""

    def __init__(
        self, did: DID, id_base: str, *, methods: VerificationRelationship = None
    ):
        super().__init__(
            did, id_base=id_base, methods=methods.items if methods else None
        )

    def _default_id_generator(self):
        """Default ID generator."""
        start = len(
            [
                vmethod
                for vmethod in self.methods
                if isinstance(vmethod, VerificationMethod)
            ]
        )
        yield from _default_id_generator(self._id_base, start=start)

    @contextmanager
    def defaults(
        self, id_generator: Iterator[str] = None, suite: VerificationSuite = None
    ) -> Iterator["RelationshipBuilder"]:
        """Enter context with defaults set."""
        self._id_generator = id_generator or self._default_id_generator()
        self._default_suite = suite
        try:
            yield self
        finally:
            self._id_generator = None
            self._default_suite = None

    def reference(self, ref: DIDUrl):
        """Add reference to relationship."""
        if not isinstance(ref, DIDUrl):
            raise ValueError(
                "Reference must be DIDUrl, not {}".format(type(ref).__name__)
            )

        self.methods.append(ref)

    def embed(
        self,
        material: Any,
        suite: VerificationSuite = None,
        ident: str = None,
        controller: DID = None,
    ):
        """Embed verification method in relationship."""
        return super().add(material, suite, ident, controller)

    def remove(self, vmethod: Union[DIDUrl, VerificationMethod]):
        """Remove reference or method from builder."""
        self.methods.remove(vmethod)


class ServiceBuilder:
    """Builder for services."""

    def __init__(self, did: DID, *, services: List[Service] = None):
        self._did = did
        self.services = services or []
        self._id_generator = None

    def _determine_next_priority(self):
        """Return the next priority after the highest priority currently in services."""
        return (
            max(
                [
                    service.priority
                    for service in self.services
                    if isinstance(service, DIDCommService)
                ]
            )
            + 1
            if self.services
            else 0
        )

    @contextmanager
    def defaults(
        self, id_generator: Iterator[str] = None
    ) -> Iterator["ServiceBuilder"]:
        """Enter context with defaults."""
        self._id_generator = id_generator or _default_id_generator(
            "service", start=len(self.services)
        )
        try:
            yield self
        finally:
            self._id_generator = None

    def add(self, type_: str, endpoint: str, ident: str = None, **extra):
        """Add service."""
        ident = ident or next(self._id_generator)
        service = Service(self._did.ref(ident), type_, endpoint, **extra)
        self.services.append(service)
        return service

    def add_didcomm(
        self,
        endpoint: str,
        recipient_keys: List[VerificationMethod],
        routing_keys: List[VerificationMethod] = None,
        *,
        priority: int = None,
        type_: str = None,
        ident: str = None
    ):
        """Add DIDComm Service."""
        ident = ident or next(self._id_generator)
        recipient_keys = [vmethod.id for vmethod in recipient_keys]
        routing_keys = routing_keys or []
        routing_keys = [vmethod.id for vmethod in routing_keys]
        priority = priority or self._determine_next_priority()
        service = DIDCommService(
            self._did.ref(ident),
            endpoint,
            recipient_keys,
            routing_keys=routing_keys,
            type_=type_,
            priority=priority,
        )
        self.services.append(service)
        return service

    def remove(self, service: Service):
        """Remove service from builder."""
        self.services.remove(service)


class DIDDocumentBuilder:
    """Builder for constructing DID Documents programmatically."""

    DEFAULT_CONTEXT = ["https://www.w3.org/ns/did/v1"]

    @validate_init(id_=Switch(All(str, Coerce(DID)), DID))
    def __init__(
        self,
        id_: DID,
        context: List[str] = None,
        *,
        also_known_as: List[str] = None,
        controller: List[str] = None
    ):
        """Initliaze builder."""
        self.id = id_
        self.context = context or self.DEFAULT_CONTEXT
        self.also_known_as = also_known_as
        self.controller = controller
        self.verification_methods = VerificationMethodBuilder(self.id)
        self.authentication = RelationshipBuilder(self.id, "auth")
        self.assertion_method = RelationshipBuilder(self.id, "assert")
        self.key_agreement = RelationshipBuilder(self.id, "key-agreement")
        self.capability_invocation = RelationshipBuilder(
            self.id, "capability-invocation"
        )
        self.capability_delegation = RelationshipBuilder(
            self.id, "capability-delegation"
        )
        self.services = ServiceBuilder(self.id)
        self.extra = {}

    @classmethod
    def from_doc(cls, doc: DIDDocument) -> "DIDDocumentBuilder":
        """Create a Builder from an existing DIDDocument."""
        builder = cls(
            id_=doc.did,
            context=doc.context,
            also_known_as=doc.also_known_as,
            controller=doc.controller,
        )
        builder.verification_methods = VerificationMethodBuilder(
            doc.did, methods=doc.verification_method
        )
        builder.authentication = RelationshipBuilder(
            doc.did, "auth", methods=doc.authentication
        )
        builder.assertion_method = RelationshipBuilder(
            doc.did, "assert", methods=doc.assertion_method
        )
        builder.key_agreement = RelationshipBuilder(
            doc.did, "key-agreement", methods=doc.key_agreement
        )
        builder.capability_invocation = RelationshipBuilder(
            doc.did, "capability-invocation", methods=doc.capability_invocation
        )
        builder.capability_delegation = RelationshipBuilder(
            doc.did, "capability-delegation", methods=doc.capability_delegation
        )
        builder.services = ServiceBuilder(doc.did, services=doc.service)
        return builder

    def build(self) -> DIDDocument:
        """Build document."""
        return DIDDocument(
            id=str(self.id),
            context=self.context,
            also_known_as=self.also_known_as,
            controller=self.controller,
            verification_method=self.verification_methods.methods or None,
            authentication=VerificationRelationship(self.authentication.methods)
            or None,
            assertion_method=VerificationRelationship(self.assertion_method.methods)
            or None,
            key_agreement=VerificationRelationship(self.key_agreement.methods) or None,
            capability_invocation=VerificationRelationship(
                self.capability_invocation.methods
            )
            or None,
            capability_delegation=VerificationRelationship(
                self.capability_delegation.methods
            )
            or None,
            service=self.services.services or None,
            **self.extra
        )



"""Export options for use externally."""

from .doc.doc_options import DIDDocumentOption
from .doc.verification_method_options import VerificationMethodOptions

doc_allow_public_key = DIDDocumentOption.allow_public_key
doc_insert_missing_ids = DIDDocumentOption.insert_missing_ids
vm_allow_type_list = VerificationMethodOptions.allow_type_list
vm_allow_missing_controller = VerificationMethodOptions.allow_missing_controller
vm_allow_controller_list = VerificationMethodOptions.allow_controller_list

__all__ = [
    "doc_allow_public_key",
    "doc_insert_missing_ids",
    "vm_allow_controller_list",
    "vm_allow_missing_controller",
    "vm_allow_type_list",
]



"""Validation tools and helpers."""

from abc import ABCMeta, abstractclassmethod, abstractmethod
from collections import namedtuple
from contextlib import contextmanager
from enum import Enum, EnumMeta
from functools import wraps
from typing import Any, Iterable, Set, Type

import voluptuous
from voluptuous import ALLOW_EXTRA, Invalid, MultipleInvalid, Required, Schema


def validate_init(*s_args, **s_kwargs):
    """Wrapper around voluptuous.validate decorator for better errors."""

    def _validate_init(func):
        func = voluptuous.validate(*s_args, **s_kwargs)(func)

        @wraps(func)
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except MultipleInvalid as err:
                raise ValueError(
                    "Invalid argument for {}:\n\t{}".format(
                        func.__qualname__,
                        "\n\t".join(
                            [
                                "{}: {}".format(error.path, Exception.__str__(error))
                                for error in err.errors
                            ]
                        ),
                    )
                ) from err

        return _wrapper

    return _validate_init


@contextmanager
def wrap_validation_error(error_to_raise: Type[Exception], message: str = None):
    """Wrap voluptuous erros with more friendly errors."""
    try:
        yield
    except MultipleInvalid as error:
        raise error_to_raise(
            "{}:\n\t{}".format(
                message or "Validation error",
                "\n\t".join([str(sub_error) for sub_error in error.errors]),
            )
        ) from error


class Into:
    """Validator that always returns a given value."""

    def __init__(self, from_value, to_value):
        self.from_value = from_value
        self.to_value = to_value

    def __call__(self, value):
        """Validate."""
        if value != self.from_value:
            raise Invalid("{} does not match {}".format(value, self.from_value))
        return self.to_value


def unwrap_if_list_of_one(value):
    """Unwrap a value froma list if list is length one."""
    if value and len(value) == 1:
        return value[0]
    return value


def single_to_list(value):
    """Wrap value in list if not already."""
    if isinstance(value, list):
        return value
    return [value]


def serialize(value):
    """Call serialize on passed value."""
    return value.serialize()


class Properties:
    """Aggregator of property info for use in validation and serialization."""

    Property = namedtuple(
        "Property",
        ["name", "data_key", "required", "validate", "serialize", "deserialize"],
    )

    def __init__(self, required=None, extra=None):
        self.validator = {}
        self.serializer = {}
        self.deserializer = {}
        self.props = []
        self.extra = extra
        self.required = required

    def add(
        self,
        *,
        data_key: str = None,
        required: bool = False,
        validate: Any = None,
        serialize: Any = None,
        deserialize: Any = None
    ):
        """Decorator defining field information."""

        def _add(func):
            prop = self.Property(
                func.__name__,
                data_key,
                required,
                validate or object,
                serialize or validate or object,
                deserialize or validate or object,
            )
            self.props.append(prop)

            key = data_key or prop.name
            key = Required(key) if required else key
            self.validator[key] = prop.validate
            if data_key:
                self.serializer[Into(prop.name, data_key)] = prop.serialize
                self.deserializer[Into(data_key, prop.name)] = prop.deserialize
            else:
                self.serializer[key] = prop.serialize
                self.deserializer[key] = prop.deserialize
            return func

        return _add

    @property
    def names(self):
        """Return names of properties."""
        return [prop.name for prop in self.props]

    def validate(self, value):
        """Validate properties from value."""
        return voluptuous.Schema(
            self.validator,
            required=self.required,
            extra=self.extra or voluptuous.PREVENT_EXTRA,
        )(value)

    def serialize(self, value):
        """Serialize properties."""
        return voluptuous.Schema(self.serializer, extra=ALLOW_EXTRA)(
            {
                name: getattr(value, name)
                for name in self.names
                if getattr(value, name) is not None
            }
        )

    def deserialize(self, value):
        """Deserialize properties."""
        return voluptuous.Schema(self.deserializer, extra=ALLOW_EXTRA)(value)


class ABCEnumMeta(EnumMeta, ABCMeta):
    """Metaclass combo for Enum ABCs."""


class Option(Enum, metaclass=ABCEnumMeta):
    """Base class for serialization and validation options."""

    @property
    @abstractmethod
    def priority(self) -> int:
        """Define options application priority."""

    @property
    @abstractmethod
    def schema(self) -> Schema:
        """Define schema applied for option."""

    @abstractclassmethod
    def apply(cls, value: dict, options: Set["Option"]) -> dict:
        """Apply options to value."""

    @staticmethod
    def schemas_in_application_order(options: Set["Option"]) -> Iterable[Schema]:
        """Return schemas of options in the order they should be applied."""
        return map(
            lambda option: option.schema,
            sorted(options, key=lambda option: option.priority),
        )



"""W3C DID Representation.

DID Parsing rules derived from W3C Decentrialized Identifiers v1.0 Working Draft 18
January 2021:

    https://w3c.github.io/did-core/

"""

from typing import Dict

from voluptuous import Invalid

from .common import DID_PATTERN, DIDError
from .did_url import DIDUrl


class InvalidDIDError(DIDError, Invalid):
    """Invalid DID."""


class DID:
    """DID Representation and helpers."""

    def __init__(self, did: str):
        """Validate and parse raw DID str."""
        self._raw = did
        matched = DID_PATTERN.match(did)
        if not matched:
            raise InvalidDIDError("Unable to parse DID {}".format(did))
        self._method = matched.group(1)
        self._id = matched.group(2)

    @property
    def method(self):
        """Return the method of this DID."""
        return self._method

    @property
    def method_specific_id(self):
        """Return the method specific identifier."""
        return self._id

    def __str__(self):
        """Return string representation of DID."""
        return self._raw

    def __repr__(self):
        """Return debug representation of DID."""
        return "<DID {}>".format(self._raw)

    def __eq__(self, other):
        """Test equality."""
        if isinstance(other, str):
            return self._raw == other
        if isinstance(other, DID):
            return self._raw == other._raw

        return False

    def __hash__(self):
        """Return hash."""
        return hash(self._raw)

    def url(self, path: str = None, query: Dict[str, str] = None, fragment: str = None):
        """Return a DID URL for this DID."""
        return DIDUrl(self._raw, path, query, fragment)

    def ref(self, ident: str) -> DIDUrl:
        """Return a DID reference (URL) for use as an ID in a DID Doc section."""
        return DIDUrl(self._raw, fragment=ident)

    @classmethod
    def is_valid(cls, did: str):
        """Return if the passed string is a valid DID."""
        return DID_PATTERN.match(did)

    @classmethod
    def validate(cls, did: str):
        """Validate the given string as a DID."""
        if not cls.is_valid(did):
            raise InvalidDIDError('"{}" is not a valid DID'.format(did))
        return did



"""DID Document deserialization options."""

from typing import Set, Type, Union

from voluptuous import ALLOW_EXTRA, All, Invalid, Schema, Switch

from ..did import DID
from ..validation import Into, Option
from .verification_method_options import VerificationMethodOptions

RELATIONSHIPS = (
    "authentication",
    "assertionMethod",
    "keyAgreement",
    "capabilityInvocation",
    "capabilityDelegation",
)


def _verification_methods(schema):
    return {
        "verificationMethod": [schema],
        **{
            key: [Switch(str, schema)]
            for key in (
                "authentication",
                "assertionMethod",
                "keyAgreement",
                "capabilityInvocation",
                "capabilityDelegation",
            )
        },
    }


def _options_of_type(type_: Type, options: Set[Option]):
    return filter(lambda option: isinstance(option, type_), options)


def _insert_missing_ids(value: dict):
    if "id" not in value:
        raise Invalid("No DID found in value.")

    did = DID(value["id"])
    index = 0

    def _walk(value: Union[str, list, dict]):
        nonlocal index
        if isinstance(value, str):
            return
        if isinstance(value, list):
            for item in value:
                _walk(item)
            return
        if isinstance(value, dict):
            if "id" not in value:
                value["id"] = str(did.ref("inserted-{}".format(index)))
                index += 1
                return

    for nested in value.values():
        _walk(nested)

    return value


class DIDDocumentOption(Option):
    """Container for validation options for DID Documents.

    Aggregates options of nested objects.
    """

    allow_public_key = 0, Schema(
        {Into("publicKey", "verificationMethod"): [dict]},
        extra=ALLOW_EXTRA,
    )
    insert_missing_ids = 1, Schema(_insert_missing_ids)

    @property
    def priority(self):
        return self.value[0]

    @property
    def schema(self):
        return self.value[1]

    @classmethod
    def apply(cls, value, options: Set[Option]):
        """Apply options to value."""
        # Apply DID Doc options
        doc_options = All(
            *cls.schemas_in_application_order(_options_of_type(cls, options))
        )
        value = doc_options(value)

        # Apply verification method options.
        vm_options = Schema(
            _verification_methods(
                All(
                    *cls.schemas_in_application_order(
                        _options_of_type(VerificationMethodOptions, options)
                    )
                )
            ),
            extra=ALLOW_EXTRA,
        )
        value = vm_options(value)
        return value



"""DID Document Object."""

import json
from typing import Any, List, Set, Union

from voluptuous import ALLOW_EXTRA, All, Coerce, Switch, Url

from ..did import DID
from ..did_url import DIDUrl
from ..validation import (
    Option,
    Properties,
    serialize,
    single_to_list,
    unwrap_if_list_of_one,
    wrap_validation_error,
)
from . import DIDDocumentError
from .doc_options import DIDDocumentOption
from .service import Service
from .didcomm_service import DIDCommService
from .verification_method import VerificationMethod
from .verification_relationship import VerificationRelationship


class IdentifiedResourceMismatch(DIDDocumentError):
    """Raised when two or more of the same ID point to differing resources."""


class ResourceIDNotFound(DIDDocumentError):
    """Raised when Resource ID not found in DID Document."""


class DIDDocumentValidationError(DIDDocumentError):
    """Raised when Document validation fails."""


def _didcomm_or_service_discriminant(value, validators):
    """Return the validator to use for the service."""
    if value["type"] == "did-communication" or value["type"] == "IndyAgent":
        yield validators[0]
    else:
        yield validators[1]


class DIDDocument:
    """Representation of DID Document."""

    properties = Properties(extra=ALLOW_EXTRA)

    def __init__(
        self,
        id: str,
        context: List[Any],
        *,
        also_known_as: List[str] = None,
        controller: List[str] = None,
        verification_method: List[VerificationMethod] = None,
        authentication: VerificationRelationship = None,
        assertion_method: VerificationRelationship = None,
        key_agreement: VerificationRelationship = None,
        capability_invocation: VerificationRelationship = None,
        capability_delegation: VerificationRelationship = None,
        service: List[Service] = None,
        **extra
    ):
        """Create DIDDocument."""
        self._id = id
        self._context = context
        self._also_known_as = also_known_as
        self._controller = controller
        self._verification_method = verification_method
        self._authentication = authentication
        self._assertion_method = assertion_method
        self._key_agreement = key_agreement
        self._capability_invocation = capability_invocation
        self._capability_delegation = capability_delegation
        self._service = service
        self.extra = extra

        self._index = {}
        self._index_resources()

    def _index_resources(self):
        """Index resources by ID.

        IDs are not guaranteed to be unique within the document.
        The first instance is stored in the index and subsequent id collisions
        are checked against the original. If they do not match, an error will
        be thrown.
        """

        def _indexer(item):
            if not item:
                # Attribute isn't set
                return
            if isinstance(item, DIDUrl):
                # We don't index references
                return
            if isinstance(item, list):
                for subitem in item:
                    _indexer(subitem)
                return
            if isinstance(item, VerificationRelationship):
                for subitem in item.items:
                    _indexer(subitem)
                return

            assert isinstance(item, (VerificationMethod, Service))
            if item.id in self._index and item != self._index[item.id]:
                raise IdentifiedResourceMismatch(
                    "ID {} already found in Index and Items do not match".format(
                        item.id
                    )
                )

            self._index[item.id] = item

        for item in (
            self.verification_method,
            self.authentication,
            self.assertion_method,
            self.key_agreement,
            self.capability_invocation,
            self.capability_delegation,
            self.service,
        ):
            _indexer(item)

    @property
    @properties.add(
        data_key="@context",
        required=True,
        validate=Switch(Url(), [Url()], dict, [dict]),
        serialize=unwrap_if_list_of_one,
        deserialize=single_to_list,
    )
    def context(self):
        """Return context."""
        return self._context

    @property
    @properties.add(
        required=True,
        validate=All(str, DID.validate),
        serialize=str,
        deserialize=str,
    )
    def id(self):
        """Return id."""
        return self._id

    @property
    def did(self):
        """Return the DID representation of id."""
        return DID(self._id)

    @property
    @properties.add(data_key="alsoKnownAs", validate=[str])
    def also_known_as(self):
        """Return also_known_as."""
        return self._also_known_as

    @property
    @properties.add(
        validate=Switch(All(str, DID.validate), [DID.validate]),
        serialize=All([Coerce(str)], unwrap_if_list_of_one),
        deserialize=All(single_to_list, [Coerce(DID)]),
    )
    def controller(self):
        """Return controller."""
        return self._controller

    @property
    @properties.add(
        data_key="verificationMethod",
        validate=[VerificationMethod.validate],
        serialize=[serialize],
        deserialize=[VerificationMethod.deserialize],
    )
    def verification_method(self):
        """Return verification_method."""
        return self._verification_method

    @property
    @properties.add(
        validate=VerificationRelationship.validate,
        serialize=serialize,
        deserialize=VerificationRelationship.deserialize,
    )
    def authentication(self):
        """Return authentication."""
        return self._authentication

    @property
    @properties.add(
        data_key="assertionMethod",
        validate=VerificationRelationship.validate,
        serialize=serialize,
        deserialize=VerificationRelationship.deserialize,
    )
    def assertion_method(self):
        """Return assertion_method."""
        return self._assertion_method

    @property
    @properties.add(
        data_key="keyAgreement",
        validate=VerificationRelationship.validate,
        serialize=serialize,
        deserialize=VerificationRelationship.deserialize,
    )
    def key_agreement(self):
        """Return key_agreement."""
        return self._key_agreement

    @property
    @properties.add(
        data_key="capabilityInvocation",
        validate=VerificationRelationship.validate,
        serialize=serialize,
        deserialize=VerificationRelationship.deserialize,
    )
    def capability_invocation(self):
        """Return capability_invocation."""
        return self._capability_invocation

    @property
    @properties.add(
        data_key="capabilityDelegation",
        validate=VerificationRelationship.validate,
        serialize=serialize,
        deserialize=VerificationRelationship.deserialize,
    )
    def capability_delegation(self):
        """Return capability_delegation."""
        return self._capability_delegation

    @property
    @properties.add(
        validate=[Service.validate],
        serialize=[serialize],
        deserialize=[
            Switch(
                DIDCommService.deserialize,
                Service.deserialize,
                discriminant=_didcomm_or_service_discriminant,
            )
        ],
    )
    def service(self):
        """Return service."""
        return self._service

    def dereference(self, reference: Union[str, DIDUrl]):
        """Dereference a DID URL to a document resource."""
        if isinstance(reference, str):
            reference = DIDUrl.parse(reference)

        if reference not in self._index:
            raise ResourceIDNotFound("ID {} not found in document".format(reference))
        return self._index[reference]

    @classmethod
    @wrap_validation_error(
        DIDDocumentValidationError, message="Failed to validate DID Document"
    )
    def validate(cls, value):
        """Validate against expected schema."""
        return cls.properties.validate(value)

    @wrap_validation_error(DIDDocumentError, message="Failed to serialize DID Document")
    def serialize(self):
        """Serialize DID Document."""
        value = self.properties.serialize(self)
        return {**value, **self.extra}

    @classmethod
    @wrap_validation_error(
        DIDDocumentValidationError, message="Failed to deserialize DID Document"
    )
    def deserialize(cls, value: dict, options: Set[Option] = None):
        """Deserialize DID Document."""
        if options:
            value = DIDDocumentOption.apply(value, options)
        value = cls.validate(value)
        value = cls.properties.deserialize(value)
        return cls(**value)

    @classmethod
    @wrap_validation_error(
        DIDDocumentValidationError, message="Failed to deserialize DID Document"
    )
    def from_json(cls, value: str, options: Set[Option] = None):
        """Deserialize DID Document from JSON."""
        doc_raw: dict = json.loads(value)
        return cls.deserialize(doc_raw, options)

    def to_json(self):
        """Serialize DID Document to JSON."""
        return json.dumps(self.serialize())



"""Additional validation options for Verification Methods."""

from typing import Set, Union

from voluptuous import ALLOW_EXTRA, All, Schema, Switch, Required

from ..did_url import DIDUrl
from ..validation import Option


def _option_allow_type_list(value: Union[str, list]):
    """Provide mapping for allow type list option."""
    return value[0] if isinstance(value, list) else value


def _option_allow_controller_list(value: Union[str, list]):
    """Provide mapping for allwo controller list option."""
    return value[0] if isinstance(value, list) else value


def _option_allow_missing_controller(value: dict):
    """Derive controller value from id."""
    if "controller" not in value:
        value["controller"] = DIDUrl.parse(value["id"]).did
    return value


class VerificationMethodOptions(Option):
    """Container for validation options for VerificationMethods."""

    allow_type_list = 0, Schema(
        {"type": All(Switch(str, [str]), _option_allow_type_list)}, extra=ALLOW_EXTRA
    )
    allow_missing_controller = 1, Schema(
        All(
            {Required("id"): All(str, DIDUrl.validate)},
            _option_allow_missing_controller,
        ),
        extra=ALLOW_EXTRA,
    )
    allow_controller_list = 2, Schema(
        {"controller": All(Switch(str, [str]), _option_allow_controller_list)},
        extra=ALLOW_EXTRA,
    )

    @property
    def priority(self):
        """Return the priority for this option."""
        return self.value[0]

    @property
    def schema(self):
        """Return the schema defined by option."""
        return self.value[1]

    @classmethod
    def apply(cls, value, options: Set["VerificationMethodOptions"]):
        """Apply options to value"""
        return All(*cls.schemas_in_application_order(options))(value)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydid', 'pydid.doc']

package_data = \
{'': ['*']}

install_requires = \
['voluptuous>=0.12.1,<0.13.0']

setup_kwargs = {
    'name': 'pydid',
    'version': '0.2.5',
    'description': 'Python library for validating, constructing, and representing DIDs and DID Documents',
    'long_description': '# PyDID\n\n[![pypi release](https://img.shields.io/pypi/v/pydid)](https://pypi.org/project/pydid/)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\nPython library for validating, constructing, and representing DIDs and DID Documents.\n\n## Installation\n\nUsing a virtual environment is generally recommended:\n\n```sh\n$ python -m venv env\n$ source env/bin/activate\n```\n\nInstall with pip:\n\n```sh\n$ pip install pydid\n```\n\n## Development\n\nThis project is managed with [Poetry](https://python-poetry.org/).\n\nTo begin making code changes, clone this repo and do the following to install\ndependencies:\n\n```sh\n$ python -m venv env\n$ source env/bin/activate\n$ pip install poetry\n$ poetry install\n```\n\n\n## Contributing\n\nSee [CONTRIBUTING.md](CONTRIBUTING.md).\n',
    'author': 'Daniel Bluhm',
    'author_email': 'dbluhm@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dbluhm/pydid',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.9,<4.0.0',
}


setup(**setup_kwargs)



"""PyDID"""

from .common import DIDError
from .did import DID, InvalidDIDError
from .did_url import DIDUrl, InvalidDIDUrlError
from .doc import DIDDocumentError
from .doc.doc import DIDDocument, DIDDocumentValidationError
from .doc.builder import DIDDocumentBuilder
from .doc.verification_method import (
    VerificationMethod,
    VerificationSuite,
    VerificationMethodValidationError,
)
from .doc.service import Service, ServiceValidationError
from .doc.didcomm_service import DIDCommService

__all__ = [
    "DID",
    "DIDCommService",
    "DIDDocument",
    "DIDDocumentBuilder",
    "DIDDocumentError",
    "DIDDocumentValidationError",
    "DIDError",
    "DIDUrl",
    "InvalidDIDError",
    "InvalidDIDUrlError",
    "Service",
    "ServiceValidationError",
    "VerificationMethod",
    "VerificationMethodValidationError",
    "VerificationSuite",
]



"""DID URL Object."""
from typing import Dict, Union
from urllib.parse import parse_qsl, urlencode, urlparse

from voluptuous import Invalid

from .common import DID_URL_DID_PART_PATTERN, DIDError


class InvalidDIDUrlError(DIDError, Invalid):
    """Invalid DID."""


class DIDUrl:
    """DID URL."""

    def __init__(
        self,
        did: str,
        path: str = None,
        query: Dict[str, str] = None,
        fragment: Union[str, int] = None,
    ):
        """Initialize DID URL.

        Leading '/' of path inserted if absent.
        """
        self.did = did
        if path and not path.startswith("/"):
            path = "/" + path
        self.path = path

        self.query = query
        self.fragment = str(fragment) if fragment else None
        self.url = self._stringify()

    def _stringify(self):
        """Stringify DID URL from parts.

        Leading '/' of path inserted if absent.
        Delimiters for query and fragment will be inserted.
        """
        value = self.did
        if self.path:
            value += self.path

        if self.query:
            value += "?" + urlencode(self.query)

        if self.fragment:
            value += "#" + self.fragment

        return value

    def __str__(self):
        """Return string representation of DID URL.

        Delimiters for query and fragment will be inserted.
        """
        return self.url

    def __repr__(self):
        """Return debug representation of DID URL."""
        return "<DIDUrl {}>".format(self.url)

    def __eq__(self, other):
        """Check equality."""
        if isinstance(other, str):
            return self.url == other
        if not isinstance(other, DIDUrl):
            return False
        return self.url == other.url

    def __hash__(self):
        """Hash url string."""
        return hash(self.url)

    @classmethod
    def parse(cls, url: str):
        """Parse DID URL from string."""
        matches = DID_URL_DID_PART_PATTERN.match(url)

        if not matches:
            raise InvalidDIDUrlError("DID could not be parsed from URL {}".format(url))

        did = matches.group(1)
        _, url_component = url.split(did)

        if not url_component:
            raise InvalidDIDUrlError(
                "No path, query, or fragment found in URL {}".format(url)
            )

        parts = urlparse(url_component)
        return cls(
            did,
            parts.path or None,
            dict(parse_qsl(parts.query)) if parts.query else None,
            parts.fragment or None,
        )

    @classmethod
    def as_str(
        cls,
        did: str,
        path: str = None,
        query: Dict[str, str] = None,
        fragment: str = None,
    ):
        """Form a DID URL from parts and return the string representation."""
        return str(cls(did, path, query, fragment))

    @classmethod
    def is_valid(cls, url: str):
        """Return whether the given string is a valid DID URL."""
        try:
            cls.parse(url)
        except InvalidDIDUrlError:
            return False
        else:
            return True

    @classmethod
    def validate(cls, url: str):
        """Validate the given url as a DID URL."""
        cls.parse(url)
        return url



"""Verification Relationship"""

from typing import List, Union

from voluptuous import All, Schema, Switch

from ..did_url import DIDUrl
from .verification_method import VerificationMethod


class VerificationRelationship:
    """Verification Relationship."""

    validate = Schema(
        [
            Switch(
                All(str, DIDUrl.validate),
                VerificationMethod.validate,
            )
        ]
    )

    def __init__(self, items: List[Union[DIDUrl, VerificationMethod]]):
        """Initialize verification relationship."""
        self.items = items

    def __bool__(self):
        """Return whether this relationship is empty or not."""
        return bool(self.items)

    def __contains__(self, item: VerificationMethod):
        if not isinstance(item, (DIDUrl, VerificationMethod)):
            return False
        return item in self.items or (
            isinstance(item, VerificationMethod) and item.id in self.items
        )

    def serialize(self):
        """Serialize this relationship."""

        def _serialize_item(item):
            if isinstance(item, DIDUrl):
                return str(item)
            if isinstance(item, VerificationMethod):
                return item.serialize()
            raise ValueError(
                "Unexpected type {} in VerificationRelationship".format(type(item))
            )

        return list(map(_serialize_item, self.items))

    @classmethod
    def deserialize(cls, value: List[Union[str, dict]]):
        """Deserialize into relationship."""
        value = cls.validate(value)
        deserializer = Schema(
            [Switch(All(str, DIDUrl.parse), VerificationMethod.deserialize)]
        )
        return cls(deserializer(value))
